import argparse
import asyncio
import base64
import datetime
import functools
import json
import os
import re
import time
from email.utils import format_datetime
from pathlib import Path

import aiohttp
import jinja2
from aiohttp import web
from markupsafe import Markup

BASE_DIR = Path(__file__).parent
SITE_ID = 36348
CACHE_DURATION = 3600
FORMATS = [
    ('m3u8', 'application/vnd.apple.mpegurl', 'adaptive'),
    ('mpd', 'application/dash+xml', 'adaptive'),
    ('mp4', 'video/mp4', '1080p'),
    ('mp4', 'video/mp4', '720p'),
    ('mp4', 'video/mp4', '540p'),
    ('mp4', 'video/mp4', '360p'),
    ('mp4', 'video/mp4', '240p'),
]

env = jinja2.Environment(
    loader=jinja2.FileSystemLoader(BASE_DIR / 'templates'),
    autoescape=jinja2.select_autoescape(default=True),
    trim_blocks=True,
    lstrip_blocks=True,
)
env.globals['FORMATS'] = FORMATS
env.filters['feed_date'] = format_datetime


def release_date(data):
    if data.get('release_dates'):
        s = data['release_dates'][0]['date']
        s += 'T00:00:00Z'
    else:
        s = data['created_at']
    return datetime.datetime.fromisoformat(s)
env.filters['release'] = release_date


def relative_datetime(dt):
    delta = datetime.datetime.now(tz=datetime.UTC) - dt
    if delta.days > 365:
        return f'{delta.days // 365} years ago'
    elif delta.days > 61:
        return f'{delta.days * 12 // 365} months ago'
    elif delta.days > 14:
        return f'{delta.days // 7} weeks ago'
    else:
        return f'{delta.days} days ago'
env.filters['reldate'] = relative_datetime


def isodate(dt):
    return dt.date().isoformat()
env.filters['isodate'] = isodate


def duration(value):
    total_seconds = value['seconds']
    seconds = total_seconds % 60
    minutes = (total_seconds // 60) % 60
    hours = total_seconds // 3600
    if hours:
        return f'{hours}:{minutes:02}:{seconds:02}'
    else:
        return f'{minutes:02}:{seconds:02}'
env.filters['duration'] = duration


def co2(bytes, green_hosting=0):
    # https://sustainablewebdesign.org/estimating-digital-emissions/

    # kWh/GB
    operational_datacenter = 0.055
    operational_network = 0.059
    operational_device = 0.08
    embodied_datacenter = 0.012
    embodied_network = 0.013
    embodied_device = 0.081

    # CO2e/kWh
    grid_intensity = 494

    grams = bytes / 1_000_000_000 * grid_intensity * sum([
        operational_datacenter * (1 - green_hosting),
        operational_network,
        operational_device,
        embodied_datacenter,
        embodied_network,
        embodied_device,
    ])
    return f'{grams:.3g}g CO₂'
env.filters['co2'] = co2


def linebreaks(text):
    return Markup('').join(
        Markup('<p>{}</p>').format(
            Markup('<br>').join(p.split('\n'))
        )
        for p in re.split(r'\n\n+', text)
    )
env.filters['linebreaks'] = linebreaks


class Auth:
    def __init__(self, email=None, password=None):
        self.email = email
        self.password = password
        self.future = None
        self.exp = 0

    def jwt_parse(self, s):
        raw = s.split('.')[1]
        for i in range(4):
            try:
                return json.loads(base64.urlsafe_b64decode(raw + '=' * i))
            except Exception:
                pass
        raise ValueError(s)

    async def auth(self):
        print(f'fetch subscription token for {self.email}')
        async with aiohttp.ClientSession() as session:
            async with session.get('https://www.dropout.tv/login') as response:
                response.raise_for_status()
                html = await response.text()
                authenticity_token = re.search(
                    r'name="authenticity_token" value="([^"]+)', html
                ).group(1)
            async with session.post('https://www.dropout.tv/login', data={
                'email': self.email,
                'password': self.password,
                'authenticity_token': authenticity_token,
            }) as response:
                html = await response.text()
                return re.search(r'token: "([^"]+)', html).group(1)

    async def unauth(self):
        print('fetch non-subscription token')
        async with aiohttp.ClientSession() as session:
            async with session.get('https://www.dropout.tv/browse') as response:
                html = await response.text()
                return re.search(r'token: "([^"]+)', html).group(1)

    async def get_token(self, timeout=300):
        if not self.future or (self.future.done() and self.exp < time.time() + timeout):
            self.future = asyncio.Future()
            try:
                if self.password:
                    token = await self.auth()
                else:
                    token = await self.unauth()
                parsed = self.jwt_parse(token)
                self.exp = parsed['exp']
                self.future.set_result(token)
            except Exception as e:
                self.future.set_exception(e)
                raise
        return await self.future

auth = Auth(os.getenv('DROPOUT_EMAIL'), os.getenv('DROPOUT_PASSWORD'))
unauth = Auth()


def async_cache(func):
    cache = {}
    async def wrapper(*args, **kwargs):
        key = functools._make_key(args, kwargs, typed=False)
        if key not in cache or cache[key][0] < time.time():
            future = asyncio.Future()
            cache[key] = (time.time() + CACHE_DURATION, future)
            try:
                future.set_result(await func(*args, **kwargs))
            except Exception as e:
                future.set_exception(e)
                del cache[key]
                raise
        return await cache[key][1]
    return wrapper


@async_cache
async def fetch(url, *, need_auth=False, **params):
    print('fetch', url)

    async with aiohttp.ClientSession() as session:
        if need_auth:
            token = await auth.get_token()
        else:
            token = await unauth.get_token()
        async with session.get(
            url,
            headers={'Authorization': f'Bearer {token}'},
            params=params,
        ) as response:
            if response.status == 404:
                raise web.HTTPNotFound
            response.raise_for_status()
            return await response.json()


async def fetch_items(url, **params):
    items = await fetch(url, **params)
    return items['_embedded']['items']


def render_response(request, template, **data):
    if request.query.get('json'):
        return web.json_response(data)
    tpl = env.get_template(template)
    body = tpl.render(query=request.query, **data)
    return web.Response(body=body, content_type='text/html')


def render_feed(request, col, items):
    if not env.globals.get('ORIGIN'):
        raise web.HTTPNotFound

    quality = request.query.get('quality', '540p')
    format = request.query.get('format', 'mp4')
    types = [t for f, t, q in FORMATS if q == quality and f == format]
    if not types:
        raise web.HTTPNotFound

    tpl = env.get_template('feed.xml')
    body = tpl.render(
        col=col,
        items=items,
        request=request,
        quality=quality,
        format=format,
        type=types[0],
    )
    return web.Response(body=body, content_type='text/xml')


async def home_view(request):
    sections = await fetch_items(
        'https://api.vhx.tv/browse',
        product='https://api.vhx.tv/products/28599',
        site_id=SITE_ID,
        per_page=10,
    )
    section_items = await asyncio.gather(*[
        fetch_items(sec['_links']['items']['href']) for sec in sections
    ])
    data = [
        {**sec, 'items': items} for sec, items in zip(sections, section_items)
    ]
    return render_response(request, 'home.html', sections=data)


async def search_view(request):
    if not request.query.get('q'):
        return render_response(request, 'search.html')

    data = await fetch(
        f'https://api.vhx.com/v2/sites/{SITE_ID}/search',
        q=request.query['q'],
        type='video,collection',
        per_page=50,
    )
    items = [item['entity'] for item in data['results']]
    return render_response(request, 'search.html', items=items)


async def collection_view(request):
    id = request.match_info['id']
    col, items = await asyncio.gather(
        fetch(f'https://api.vhx.tv/collections/{id}'),
        fetch_items(f'https://api.vhx.tv/collections/{id}/items'),
    )
    return render_response(request, 'collection.html', col=col, items=items)


async def collection_feed(request):
    id = request.match_info['id']
    data, items = await asyncio.gather(
        fetch(f'https://api.vhx.tv/collections/{id}'),
        fetch_items(f'https://api.vhx.tv/collections/{id}/items'),
    )
    return render_feed(request, data, items)


async def get_series(id):
    series, _seasons = await asyncio.gather(
        fetch(f'https://api.vhx.tv/collections/{id}'),
        fetch_items(f'https://api.vhx.tv/collections/{id}/items'),
    )
    season_episodes = await asyncio.gather(*[
        fetch_items(season['_links']['episodes']['href']) for season in _seasons
    ])
    seasons = [
        {**season, 'episodes': episodes}
        for season, episodes in zip(_seasons, season_episodes)
    ]
    return series, seasons


async def series_view(request):
    series, seasons = await get_series(request.match_info['id'])
    return render_response(request, 'series.html', col=series, seasons=seasons)


async def series_feed(request):
    series, seasons = await get_series(request.match_info['id'])
    items = [e for season in seasons for e in season['episodes']]
    return render_feed(request, series, items)


async def get_next(data):
    try:
        season_id = data['canonical_collection']['id']
        episode_number = data['episode_number']
        season = await fetch_items(f'https://api.vhx.tv/collections/{season_id}/items')
        if episode_number < len(season):
            return season[episode_number]['id']

        series_id = data['canonical_collection']['parent']['id']
        season_number = data['season_number']
        series = await fetch_items(f'https://api.vhx.tv/collections/{series_id}/items')
        if season_number < len(series):
            season_id = series[season_number]['id']
            season = await fetch_items(f'https://api.vhx.tv/collections/{season_id}/items')
            return season[0]['id']
    except Exception:
        return None


async def video_view(request):
    id = request.match_info['id']
    video = await fetch(f'https://api.vhx.tv/videos/{id}')
    next = await get_next(video)
    return render_response(request, 'video.html', video=video, next=next)


async def file_view(request):
    id = request.match_info['id']
    quality = request.match_info['quality']
    format = request.match_info['format']
    files = await fetch(f'https://api.vhx.tv/videos/{id}/files', need_auth=True)
    filtered = [file for file in files if file['format'] == format]
    if not filtered:
        raise web.HTTPNotFound
    if quality == 'adaptive':
        best = filtered[0]
    else:
        q = int(quality[:-1], 10)
        best = min(filtered, key=lambda f: abs(int(f['quality'][:-1], 10) - q))
    raise web.HTTPSeeOther(location=best['_links']['source']['href'])


async def favicon(request):
    raise web.HTTPSeeOther(location='/static/favicon.ico')


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--port', type=int, default=8000)
    parser.add_argument('--origin')
    args = parser.parse_args()

    env.globals['ORIGIN'] = args.origin

    app = web.Application(middlewares=[
        web.normalize_path_middleware(),
    ])
    app.router.add_static('/static', BASE_DIR / 'static')
    app.router.add_get('/favicon.ico', favicon)
    app.router.add_get('/', home_view)
    app.router.add_get('/search/', search_view)
    app.router.add_get(r'/movie/{id:\d+}/', collection_view)
    app.router.add_get(r'/playlist/{id:\d+}/', collection_view)
    app.router.add_get(r'/playlist/{id:\d+}.rss', collection_feed)
    app.router.add_get(r'/series/{id:\d+}/', series_view)
    app.router.add_get(r'/series/{id:\d+}.rss', series_feed)
    app.router.add_get(r'/video/{id:\d+}/', video_view)
    app.router.add_get(r'/video/{id:\d+}-{quality:\d+p}.{format:mp4}', file_view)
    app.router.add_get(r'/video/{id:\d+}-{quality:adaptive}.{format:(m3u8|mpd)}', file_view)  # noqa
    web.run_app(app, host='localhost', port=args.port)
